## ----------------------------------------------------------------------------
# Rishu::Remote
# -----------------------------------------------------------------------------
# 里珠・外部のプロセスの受け側.
# -----------------------------------------------------------------------------
# Mastering programed by YAMASHINA Hio
#
# Copyright 2002 YMIRLINK,Inc.
# -----------------------------------------------------------------------------
# $Id: Remote.pm,v 1.5 2002/11/09 17:35:45 hio Exp $
# -----------------------------------------------------------------------------
package Rishu::Remote;
use strict;

# -----------------------------------------------------------------------------
# Rishu::Remote->new
sub new
{
  my $pkg = shift;
  my $this = bless{},$pkg;
  
  my ($rd,$wd) = (\*STDIN,\*STDOUT);
  $wd ||= $rd;
  select((select($wd),$|=1)[0]);
  
  $this->{rd} = $rd;
  $this->{wd} = $wd;
  binmode($rd);
  binmode($wd);
  
  $this->{on_load}    = \&main::on_load;
  $this->{on_request} = \&main::on_request;
  $this->{on_unload}  = \&main::on_unload;
  
  $this;
}

# -----------------------------------------------------------------------------
# Rishu::Remote->run()
sub run
{
  my $this = shift;
  
  my $rd = $this->{rd};
  $this->log("sysmtem","run\n");
  for(;;)
  {
    # コマンド行の認識
    $_ = <$rd>;
    if( !defined($_) )
    {
      last;
    }
    s/\r?\n//;
    $this->log("recv","$_\n");
    next if( $_ eq '' );
    my $head = $_;
    my ($method,$type,$proto,$ver) = m,^(PROXY)\s+(\S+)\s+([^\s/]+)/(\d+\.\d+)$,i;
    
    # ヘッダ部分の認識
    my $in = '';
    while(<$rd>)
    {
      s/\r?\n//;
      $in .= $_."\r\n";
      if( $_ eq '' )
      {
	last;
      }
    }
    if( !$type )
    {
      $this->invalid_request("line: $head");
      next;
    }
    $type = uc($type);
    if( $type eq 'EXIT' )
    { # そのまま終了.
      last;
    }elsif( $type eq 'LOAD' )
    {
      my $dir = join '',grep{s/^Directory: //} split(/\r?\n/,$in);
      if( $dir ne '' )
      {
        my $res = $this->{on_load}->($dir);
	my $reply;
	if( $res )
	{
	  $reply = "RISHU/1.1 200 OK\r\n";
	  $reply .= "Sender: Rishu/Proxy\r\n";
	}else
	{
	  $reply = "RISHU/1.1 500 Internal Server Error\r\n";
	  $reply .= "Sender: Rishu/Proxy\r\n";
	}
	if( !defined($res) || length($res)==0 )
	{
	  $reply .= "\r\n";
        }else
	{
	  $reply .= "Content-Length: ".length($res)."\r\n";
	  $reply .= "\r\n";
	  $reply .= $res;
	}
        $this->_print($reply);
      }else
      {
	$this->invalid_request("LOAD, no Directory: element");
      }
    }elsif( $type eq 'REQUEST' )
    {
      $in =~ s/^Command:\s+//;
      my $res = $this->{on_request}->($in);
      $this->_print("RISHU/1.1 200 OK\r\nOriginal-Response: ".join("\r\n",split(/\r?\n/,$res),"\r\n"));
    }elsif( $type eq 'UNLOAD' )
    {
      my $res = $this->{on_unload}->($in);
      my $reply;
      if( $res )
      {
	$reply = "RISHU/1.1 200 OK\r\n";
	$reply .= "Sender: Rishu/Proxy\r\n";
      }else
      {
	$reply = "RISHU/1.1 500 Internal Server Error\r\n";
	$reply .= "Sender: Rishu/Proxy\r\n";
      }
      if( !defined($res) || length($res)==0 )
      {
	$reply .= "\r\n";
      }else
      {
	$reply .= "Content-Length: ".length($res)."\r\n";
	$reply .= "\r\n";
	$reply .= $res;
      }
      $this->_print($reply);
    }else
    {
      $this->invalid_request("type: $type");
    }
  }
  
  $this;
}

# -----------------------------------------------------------------------------
# $remote->invalid_request
sub invalid_request
{
  my $this = shift;
  my $cause = shift;
  $this->_print("invalid request : $cause\r\n");
  $this;
}

# -----------------------------------------------------------------------------
# $remote->_print
sub _print
{
  my $this = shift;
  my $sym = $this->{wd};
  print $sym @_;
  $this;
}

# -----------------------------------------------------------------------------
# $remote->log('type','text');
sub log
{
  my $this = shift;
  my $type = shift;
  my $text = shift;
  
  # not implemented yet.
  
  $this;
}

# -----------------------------------------------------------------------------
# end of package
# -----------------------------------------------------------------------------
1;
