/* ----------------------------------------------------------------------------
 * rishu_proxy.cpp
 *   里珠/Proxy の栞のみでの実装.
 * ----------------------------------------------------------------------------
 * Mastering programed by YAMASHINA Hio
 *
 * Copyright 2002 YMIRLINK,Inc.
 * ----------------------------------------------------------------------------
 * $Id: rishu_proxy.c,v 1.8 2002/11/10 09:16:31 hio Exp $
 * ----------------------------------------------------------------------------
 */

#include <windows.h>
#include "rishu_proxy.h"
#include <stdio.h>
#include <string.h>
#include <tchar.h>

/* ----------------------------------------------------------------------------
 * LPTSTR message = getLastErrorMessage();
 * LPTSTR message = getErrorMessage(DWORD errorCode);
 *   エラーメッセージの取得 
 *   取得したメッセージは LocalFree で解放してね☆ 
 */
LPTSTR getLastErrorMessage(void)
{
  return getErrorMessage(GetLastError());
}
LPTSTR getErrorMessage(DWORD errcode)
{
  LPVOID lpMessage;
  FormatMessage( FORMAT_MESSAGE_ALLOCATE_BUFFER
		 | FORMAT_MESSAGE_FROM_SYSTEM
		 | FORMAT_MESSAGE_IGNORE_INSERTS,
		 NULL,
		 errcode,
		 MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), /* 既定の言語 */
		 (LPTSTR)&lpMessage,
		 0,
		 NULL
		 );
  return lpMessage;
}

/* ----------------------------------------------------------------------------
 * ANSI => Unicode 変換
 *   result: 書き込まれた文字数(TCHARの数).
 *           エラー時 0.
 *           nOutStringLen==0 だと必要な文字数.
 */
int
AnsiToUnicode(LPWSTR lpszOutputString,
	      int    nOutStringLen,
	      LPCSTR lpInputString,
	      int    nInStringLen
	      )
{
  return MultiByteToWideChar(CP_ACP,
			     0,
			     lpInputString, nInStringLen,
			     lpszOutputString, nOutStringLen);
}

/* ----------------------------------------------------------------------------
 * Unicode => ANSI 変換.
 */
int
UnicodeToAnsi(LPSTR   lpszOutputString,
	      int     nOutStringLen,
	      LPCWSTR lpInputString,
	      int     nInStringLen
	      )
{
  return WideCharToMultiByte(CP_ACP,
			     0,
			     lpInputString, nInStringLen,
			     lpszOutputString, nOutStringLen,
			     NULL, NULL
			     );
}


/* ----------------------------------------------------------------------------
 * 里珠/Proxy の初期化.
 */
RishuProxy* RishuProxy_init(RishuProxy* proxy,
			    WORD wSize,
			    HANDLE hDll
			    )
{
  if( wSize<sizeof(RishuProxy) )
  {
    return NULL;
  }
  proxy->wSize = wSize;
  proxy->wVersion = 0x0101;
  
  proxy->hDll = hDll;
  proxy->hGlobal_loaddir = NULL;
  proxy->loaddir = NULL;
  proxy->dwLoaddirLen = 0;
  
  proxy->hChlProcess = NULL;
  proxy->hChlInput = NULL;
  proxy->hChlOutput = NULL;
  
  proxy->dwLogBits =
    (1<<RishuProxy_LOG_FATAL)
    | (1<<RishuProxy_LOG_ERROR)
    | (1<<RishuProxy_LOG_SYSTEM)
    | (1<<RishuProxy_LOG_CLEANUP)
    ;
  proxy->logfilename = NULL;
  
  proxy->msgseq = 0;
  proxy->readbuf = NULL;
  proxy->dwReadBufLen = 0;
  proxy->dwReadBufSize = 0;
  
  proxy->hHeap = NULL;
  
  
  /* ヒープオブジェクトの作成 */
  proxy->hHeap = HeapCreate(HEAP_NO_SERIALIZE,64*1024,0);
  if( proxy->hHeap==NULL )
  {
    return NULL;
  }
  
  /* ini ファイルの確認. */
  {
    LPTSTR filename;
    DWORD  filename_charsize;
    DWORD  len;
    DWORD  tail;
    
    DWORD dwRes;
    LPTSTR tmpbuf;
    DWORD tmpbuf_charsize;
    DWORD i;
      
    /* ファイル名の生成 */
    filename_charsize = 512;
    filename = (LPTSTR)RishuProxy_async_myAlloc(proxy,filename_charsize*sizeof(TCHAR));
    if( !filename )
    {
      RishuProxy_discard(proxy);
      return NULL;
    }
    filename_charsize = RishuProxy_async_mySize(proxy,filename);
    for(;;)
    {
      len = GetModuleFileName(proxy->hDll,filename,filename_charsize-1);
      if( len==0 )
      {
	/* GetModuleFileName failed. */
	RishuProxy_async_myFree(proxy,filename);
	RishuProxy_discard(proxy);
	return NULL;
      }
      if( len<filename_charsize-2 )
      {
	break;
      }
      filename_charsize *= 2;
      filename = (LPTSTR)RishuProxy_async_myReAlloc(proxy,filename,filename_charsize*sizeof(TCHAR));
      if( !filename )
      {
	RishuProxy_discard(proxy);
	return NULL;
      }
      filename_charsize = RishuProxy_async_mySize(proxy,filename);
    }
    
    /* 拡張子辺りの書き換え */
    tail = len;
    for(;;)
    {
      if( tail==0 || filename[--tail]==TEXT('\\') )
      {
	tail = len;
	break;
      }
      if( filename[tail]==TEXT('.') )
      {
	break;
      }
    }
    if( tail+4+1>=filename_charsize )
    {
      filename_charsize = tail+4+1;
      filename = (LPTSTR)RishuProxy_async_myReAlloc(proxy,filename,filename_charsize*sizeof(TCHAR));
      if( !filename )
      {
	RishuProxy_discard(proxy);
	return NULL;
      }
      filename_charsize = RishuProxy_async_mySize(proxy,filename);
    }
    _tcsncpy(filename+tail,TEXT(".ini"),5);
    
    /* ログファイル名も生成〜. */
    proxy->logfilename  = (LPTSTR)RishuProxy_async_myAlloc(proxy,tail+9);
    if( !proxy->logfilename )
    {
      RishuProxy_discard(proxy);
      return NULL;
    }
    CopyMemory(proxy->logfilename,filename,tail*sizeof(TCHAR));
    CopyMemory(proxy->logfilename+tail,TEXT("_log.txt"),9*sizeof(TCHAR));
    RishuProxy_log(proxy, TEXT("system"),
		   TEXT("RishuProxy::init"),
		   TEXT("now initializing ...\n")
		   );
    
    /* 設定の読み込み. */
    tmpbuf_charsize = 512;
    tmpbuf = (LPTSTR)RishuProxy_async_myAlloc(proxy,tmpbuf_charsize*sizeof(TCHAR));

    /* ログ設定 */
    RishuProxy_log(proxy, TEXT("system"),
		   TEXT("RishuProxy::init"),
		   TEXT("loading ini file ... [%s]\n"),
		   filename
		   );
    for(;;)
    {
      dwRes = GetPrivateProfileString(TEXT("Rishu/Proxy"),
				      TEXT("log"),
				      TEXT("nya"),
				      tmpbuf,
				      tmpbuf_charsize,
				      filename
				      );
      if( dwRes!=tmpbuf_charsize-1 )
      {
	break;
      }
      tmpbuf_charsize *= 2;
      tmpbuf = (LPTSTR)RishuProxy_async_myAlloc(proxy,tmpbuf_charsize*sizeof(TCHAR));
      if( !tmpbuf )
      {
	RishuProxy_discard(proxy);
	return NULL;
      }
    }
    
    for( i=0; i<dwRes; ++i )
    {
      DWORD st;
      BOOL  on;
      DWORD bit;
      if( _istspace(tmpbuf[i]) )
      {
	continue;
      }
      if( tmpbuf[i]==TEXT('-') )
      {
	on = FALSE;
	++i;
      }else if( tmpbuf[i]==TEXT('+') )
      {
	on = TRUE;
	++i;
      }else
      {
	on = TRUE;
      }
      st = i;
      while( ++i<dwRes && !_istspace(tmpbuf[i]) )
      {
      }
      if( i-st==5 && _tcsnicmp(tmpbuf+st,TEXT("fatal"),5)==0 )
      {
	bit = 1<<RishuProxy_LOG_FATAL;
      }else if( i-st==5 && _tcsnicmp(tmpbuf+st,TEXT("error"),5)==0 )
      {
	bit = 1<<RishuProxy_LOG_ERROR;
      }else if( i-st==4 && _tcsnicmp(tmpbuf+st,TEXT("warn"),4)==0 )
      {
	bit = 1<<RishuProxy_LOG_WARN;
      }else if( i-st==6 && _tcsnicmp(tmpbuf+st,TEXT("system"),6)==0 )
      {
	bit = 1<<RishuProxy_LOG_SYSTEM;
	on = TRUE;
      }else if( i-st==5 && _tcsnicmp(tmpbuf+st,TEXT("trace"),5)==0 )
      {
	bit = 1<<RishuProxy_LOG_TRACE;
      }else if( i-st==5 && _tcsnicmp(tmpbuf+st,TEXT("debug"),5)==0 )
      {
	bit = 1<<RishuProxy_LOG_DEBUG;
      }else
      {
	RishuProxy_log(proxy, TEXT("system"),
		       TEXT("RishuProxy::init"),
		       TEXT("[unknown logtype] [[%.*s]]\n"),
		       i-st, tmpbuf+st
		       );
	continue;
      }
      if( on )
      {
	proxy->dwLogBits |= bit;
      }else
      {
	proxy->dwLogBits &= ~bit;
      }
    }
  }
  proxy->dwLogBits |= 1<<RishuProxy_LOG_SYSTEM;
  if( RishuProxy_tracelogIsEnabled(proxy) )
  {
    RishuProxy_log(proxy, TEXT("trace"),
		   TEXT("RishuProxy::init"),
		   TEXT("initialize done\n"
			"[[\nlogbits==%08x\n]]\n"),
		   proxy->dwLogBits
		   );
  }
  
  return proxy;
}

/* ----------------------------------------------------------------------------
 * 里珠/Proxy の破棄.
 */
void RishuProxy_discard(RishuProxy* proxy)
{
  if( proxy->loaddir )
  {
    if( proxy->loaddir!=(LPTSTR)proxy->hGlobal_loaddir )
    {
      RishuProxy_myFree(proxy,proxy->loaddir);
    }
    proxy->loaddir = NULL;
    proxy->dwLoaddirLen = 0;
  }
  if( proxy->hGlobal_loaddir )
  {
    GlobalFree(proxy->hGlobal_loaddir);
    proxy->hGlobal_loaddir = NULL;
    proxy->dwLoaddirLen = 0;
  }
  if( proxy->hChlProcess )
  {
    TerminateProcess(proxy->hChlProcess,1);
    CloseHandle(proxy->hChlProcess);
    proxy->hChlProcess = NULL;
  }
  if( proxy->hChlInput )
  {
    CloseHandle(proxy->hChlInput);
    proxy->hChlInput = NULL;
  }
  if( proxy->hChlOutput )
  {
    CloseHandle(proxy->hChlOutput);
    proxy->hChlOutput = NULL;
  }
  if( proxy->readbuf )
  {
    RishuProxy_async_myFree(proxy,proxy->readbuf);
    proxy->readbuf = NULL;
    proxy->dwReadBufLen = 0;
    proxy->dwReadBufSize = 0;
  }
  if( proxy->hHeap )
  {
    BOOL res = HeapDestroy(proxy->hHeap);
    proxy->hHeap = NULL;
  }
}

/* ----------------------------------------------------------------------------
 * 排他制御 - lock
 */
void RishuProxy_lock(RishuProxy* proxy)
{
  /* 未実装, not implemented yet */
  return;
}

/* ----------------------------------------------------------------------------
 * 排他制御 - trylock
 */
BOOL RishuProxy_trylock(RishuProxy* proxy)
{
  /* 未実装, not implemented yet */
  return TRUE;
}

/* ----------------------------------------------------------------------------
 * 排他制御 - unlock
 */
void RishuProxy_unlock(RishuProxy* proxy)
{
  /* 未実装, not implemented yet */
  return;
}

/* ---------------------------------------------------------------------------
 * ヒープ操作 - alloc
 */
LPVOID RishuProxy_async_myAlloc(RishuProxy* proxy,
				SIZE_T size
				)
{
  LPVOID ptr;
  
  if( !proxy )
  {
    return NULL;
  }
  
  /* メモリの確保 
   *   (失敗しても SetLastError はされない)
   */
  ptr = HeapAlloc(proxy->hHeap, HEAP_NO_SERIALIZE, size);

  return ptr;
}

/* ---------------------------------------------------------------------------
 * ヒープ操作 - realloc
 */
LPVOID RishuProxy_async_myReAlloc(RishuProxy* proxy,
				  LPVOID ptr,
				  DWORD size
				  )
{
  if( !proxy )
  {
    return NULL;
  }
  if( !proxy->hHeap )
  {
    if( RishuProxy_fatallogIsEnabled(proxy) )
    {
      RishuProxy_log(proxy, TEXT("fatal"),
		     TEXT("RishuProxy::async_myReAlloc"),
		     TEXT("!! no heap created yet !!\n")
		     );
    }
    return NULL;
  }
  
  return HeapReAlloc(proxy->hHeap,HEAP_NO_SERIALIZE,ptr,size);
}

/* ---------------------------------------------------------------------------
 * ヒープ操作 - free
 */
void RishuProxy_async_myFree(RishuProxy* proxy,
			     LPVOID ptr
			     )
{
  if( !proxy )
  {
    return;
  }
  if( !proxy->hHeap )
  {
    if( RishuProxy_fatallogIsEnabled(proxy) )
    {
      RishuProxy_log(proxy, TEXT("fatal"),
		     TEXT("RishuProxy::async_myFree"),
		     TEXT("!! no heap created yet !!\n")
		     );
    }
    return;
  }
  HeapFree(proxy->hHeap,HEAP_NO_SERIALIZE,ptr);
  return;
}

/* ---------------------------------------------------------------------------
 * ヒープ操作 - size
 */
SIZE_T RishuProxy_async_mySize(RishuProxy* proxy,
			       LPVOID ptr
			       )
{
  if( !proxy )
  {
    return 0;
  }
  if( !proxy->hHeap )
  {
    if( RishuProxy_fatallogIsEnabled(proxy) )
    {
      RishuProxy_log(proxy, TEXT("fatal"),
		     TEXT("RishuProxy::async_mySize"),
		     TEXT("!! no heap created yet !!\n")
		     );
    }
    return 0;
  }
  return HeapSize(proxy->hHeap,HEAP_NO_SERIALIZE,ptr);
}

/* ---------------------------------------------------------------------------
 * 同期付きヒープ操作 - alloc
 */
LPVOID RishuProxy_myAlloc(RishuProxy* proxy,
			  SIZE_T size
			  )
{
  LPVOID ptr;
  RishuProxy_lock(proxy);
  ptr = RishuProxy_async_myAlloc(proxy,size);
  RishuProxy_unlock(proxy);
  return ptr;
}

/* ---------------------------------------------------------------------------
 * 同期付きヒープ操作 - realloc
 */
LPVOID RishuProxy_myReAlloc(RishuProxy* proxy,
			    LPVOID ptr,
			    SIZE_T size
			    )
{
  RishuProxy_lock(proxy);
  ptr = RishuProxy_async_myReAlloc(proxy,ptr,size);
  RishuProxy_unlock(proxy);
  return ptr;
}

/* ---------------------------------------------------------------------------
 * 同期付きヒープ操作 - free
 */
void RishuProxy_myFree(RishuProxy* proxy,
		       LPVOID ptr
		       )
{
  RishuProxy_lock(proxy);
  RishuProxy_async_myFree(proxy,ptr);
  RishuProxy_unlock(proxy);
  return;
}

/* ---------------------------------------------------------------------------
 * 同期付きヒープ操作 - size
 */
extern SIZE_T RishuProxy_mySize(RishuProxy* proxy,
				LPVOID ptr
				)
{
  SIZE_T size;
  RishuProxy_lock(proxy);
  size = RishuProxy_async_mySize(proxy,ptr);
  RishuProxy_unlock(proxy);
  return size;
}


/* ----------------------------------------------------------------------------
 * loaddirからの相対ディレクトリを, loaddir を含めたものに変換.
 */
LPTSTR RishuProxy_getPath(RishuProxy* proxy,
			  LPCTSTR relativePath,
			  DWORD relativePath_len
			  )
{
  LPTSTR expanded;
  
  if( !proxy )
  {
    return NULL;
  }
  
  if( !proxy->loaddir )
  {
    if( !proxy->hGlobal_loaddir )
    {
      if( RishuProxy_errorlogIsEnabled(proxy) )
      {
	RishuProxy_log(proxy, TEXT("error"),
		       TEXT("RishuProxy::getPath"),
		       TEXT("on_load was not called yet.\n")
		       );
      }
    }
#ifndef UNICODE
    /* ANSI */
    proxy->loaddir = (LPSTR) proxy->hGlobal_loaddir;
#else
    /* UNICODE */
#error not suspported yet.
#endif /* defined(UNICODE) */
  }
  
  expanded = (LPTSTR)RishuProxy_myAlloc(proxy,(proxy->dwLoaddirLen+relativePath_len+1)*sizeof(TCHAR));
  if( !expanded )
  {
    if( RishuProxy_errorlogIsEnabled(proxy) )
    {
      RishuProxy_log(proxy, TEXT("error"),
		     TEXT("RishuProxy::getPath"),
		     TEXT("proxy->myAlloc failed\n")
		     );
    }
    return NULL;
  }
  CopyMemory(expanded,
	     proxy->loaddir,
	     proxy->dwLoaddirLen*sizeof(TCHAR)
	     );
  CopyMemory(expanded+proxy->dwLoaddirLen,
	     relativePath,
	     (relativePath_len+1)*sizeof(TCHAR)
	     );
  
  return expanded;
}


/* ----------------------------------------------------------------------------
 * 栞 Method / load
 */ 
BOOL RishuProxy_on_load(RishuProxy* proxy,
			HGLOBAL hGlobal_loaddir,
			long loaddir_len
			)
{
  BOOL res;
  
  HANDLE par_wd,par_rd;
  HANDLE chl_wd,chl_rd;
  HANDLE chl_wd_dup,chl_rd_dup;
  HANDLE hCurrentProcess;

  LPTSTR commandLine;
  DWORD dwCreationFlags;
  STARTUPINFO startupInfo;
  PROCESS_INFORMATION processInfo;
  
  RishuProxy_Request request;
  RishuProxy_Response response;
  RishuProxy_Response* ptr_response;
  
  if( !proxy )
  {
    return FALSE;
  }

  if( RishuProxy_tracelogIsEnabled(proxy) )
  {
    RishuProxy_log(proxy, TEXT("trace"),
		   TEXT("RishuProxy::on_load"),
		   TEXT("enter\n")
		   );
  }
  
  /* --------------------------------------------------------------------------
   * 接続中のプロセスがあったら終了しておく (あるはずないけど) 
   */
  if( proxy->hChlProcess )
  {
    if( RishuProxy_warnlogIsEnabled(proxy) )
    {
      RishuProxy_log(proxy, TEXT("warn"),
		     TEXT("RishuProxy::on_load"),
		     TEXT("remote process is already running with pid 0x%08x, do terminate\n"),
		     proxy->hChlProcess
		     );
    }
    res = TerminateProcess(proxy->hChlProcess,1);
    proxy->hChlProcess = NULL;
    if( !res && RishuProxy_warnlogIsEnabled(proxy) )
    {
      LPTSTR cause = getLastErrorMessage();
      RishuProxy_log(proxy,TEXT("warn"),
		     TEXT("RishuProxy::on_load"),
		     TEXT("terminateProcess failed\n>> %s\n"),
		     cause
		     );
      LocalFree(cause);
    }
  }
  if( proxy->hChlInput )
  {
    CloseHandle(proxy->hChlInput);
    proxy->hChlInput = NULL;
  }
  if( proxy->hChlOutput )
  {
    CloseHandle(proxy->hChlOutput);
    proxy->hChlOutput = NULL;
  }
  
  /* 古いデータがあったらけしておく. */
  if( proxy->loaddir )
  {
    if( proxy->loaddir!=(LPTSTR)proxy->hGlobal_loaddir )
    {
      RishuProxy_myFree(proxy,proxy->loaddir);
    }
    proxy->loaddir = NULL;
  }
  if( proxy->hGlobal_loaddir )
  {
    GlobalFree(proxy->hGlobal_loaddir);
  }
  proxy->hGlobal_loaddir = hGlobal_loaddir;
  proxy->dwLoaddirLen = loaddir_len;
    
  /* --------------------------------------------------------------------------
   * 通信用のパイプの作成.
   */
  hCurrentProcess = GetCurrentProcess();
  if( !CreatePipe(&par_rd,&chl_wd,NULL,0) )
  {
    if( RishuProxy_errorlogIsEnabled(proxy) )
    {
      LPTSTR cause = getLastErrorMessage();
      RishuProxy_log(proxy,TEXT("error"),
		     TEXT("RishuProxy::on_load"),
		     TEXT("CreatePipe failed (1/2)\n>> %s\n"),
		     cause
		     );
      LocalFree(cause);
    }
    if( RishuProxy_tracelogIsEnabled(proxy) )
    {
      RishuProxy_log(proxy,TEXT("trace"),
		     TEXT("RishuProxy::on_load"),
		     TEXT("failure, CreatePipe (1/2)\n")
		     );
    }
    return FALSE;
  }
  if( !DuplicateHandle(hCurrentProcess, chl_wd,
		       hCurrentProcess, &chl_wd_dup,
		       0,
		       TRUE,
		       DUPLICATE_SAME_ACCESS
		       ) )
  {
    if( RishuProxy_errorlogIsEnabled(proxy) )
    {
      LPTSTR cause = getLastErrorMessage();
      RishuProxy_log(proxy,TEXT("error"),
		     TEXT("RishuProxy::on_load"),
		     TEXT("DuplicateHandle failed (1/2)\n>> %s\n"),
		     cause
		     );
      LocalFree(cause);
    }
    CloseHandle(par_rd);
    CloseHandle(chl_wd);
    if( RishuProxy_tracelogIsEnabled(proxy) )
    {
      RishuProxy_log(proxy,TEXT("trace"),
		     TEXT("RishuProxy::on_load"),
		     TEXT("failure, DuplicateHandle (1/2)\n")
		     );
    }
    return FALSE;
  }
  CloseHandle(chl_wd);
  chl_wd = NULL;
  
  if( !CreatePipe(&chl_rd,&par_wd,NULL,0) )
  {
    if( RishuProxy_errorlogIsEnabled(proxy) )
    {
      LPTSTR cause = getLastErrorMessage();
      RishuProxy_log(proxy,TEXT("error"),
		     TEXT("RishuProxy::on_load"),
		     TEXT("CreatePipe failed (2/2)\n>> %s\n"),
		     cause
		     );
      LocalFree(cause);
    }
    CloseHandle(par_rd);
    CloseHandle(chl_wd_dup);
    if( RishuProxy_tracelogIsEnabled(proxy) )
    {
      RishuProxy_log(proxy,TEXT("trace"),
		     TEXT("RishuProxy::on_load"),
		     TEXT("failure, CreatePipe (2/2)\n")
		     );
    }
    return FALSE;
  }
  if( !DuplicateHandle(hCurrentProcess, chl_rd,
		       hCurrentProcess, &chl_rd_dup,
		       0,
		       TRUE,
		       DUPLICATE_SAME_ACCESS
		       ) )
  {
    if( RishuProxy_errorlogIsEnabled(proxy) )
    {
      LPTSTR cause = getLastErrorMessage();
      RishuProxy_log(proxy,TEXT("error"),
		     TEXT("RishuProxy::on_load"),
		     TEXT("DuplicateHandle failed (2/2)\n>> %s\n"),
		     cause
		     );
      LocalFree(cause);
    }
    CloseHandle(par_rd);
    CloseHandle(par_wd);
    CloseHandle(chl_rd);
    CloseHandle(chl_wd_dup);
    if( RishuProxy_tracelogIsEnabled(proxy) )
    {
      RishuProxy_log(proxy,TEXT("trace"),
		     TEXT("RishuProxy::on_load"),
		     TEXT("failure, DuplicateHandle (2/2)\n")
		     );
    }
    return FALSE;
  }
  CloseHandle(chl_rd);
  chl_rd = NULL;
  if( RishuProxy_debuglogIsEnabled(proxy) )
  {
    RishuProxy_log(proxy, TEXT("debug"),
		   TEXT("RishuProxy::on_load"),
		   TEXT("CreatePipe ok\n")
		   );
  }
  
  /* --------------------------------------------------------------------------
   * 実行ファイルの位置.
   */
  {
    LPCTSTR relative_perl = TEXT("modules\\5.8.0\\bin\\MSWin32-x86-multi-thread\\perl.exe");
    LPCTSTR relative_script = TEXT("rishu_remote.pl");
    DWORD relative_perl_len;
    DWORD relative_script_len;
    
    LPTSTR path_perl;
    LPTSTR path_script;
    DWORD path_perl_len;
    DWORD path_script_len;
    
    /* perl へのパスを変換. */
    relative_perl_len = _tcslen(relative_perl);
    path_perl = RishuProxy_getPath(proxy,relative_perl,relative_perl_len);
    if( !path_perl )
    {
      if( RishuProxy_errorlogIsEnabled(proxy) )
      {
	RishuProxy_log(proxy, TEXT("error"),
		       TEXT("RishuProxy::on_load"),
		       TEXT("proxy->getPath failed (perl.exe)\n")
		       );
      }
      if( RishuProxy_tracelogIsEnabled(proxy) )
      {
	RishuProxy_log(proxy, TEXT("trace"),
		       TEXT("RishuProxy::on_load"),
		       TEXT("proxy->getPath failed (perl.exe)\n")
		       );
      }
      return FALSE;
    }
    
    /* script へのパスを変換. */
    relative_script_len = _tcslen(relative_script);
    path_script = RishuProxy_getPath(proxy,relative_script,relative_script_len);
    if( !path_script )
    {
      if( RishuProxy_errorlogIsEnabled(proxy) )
      {
	RishuProxy_log(proxy, TEXT("error"),
		       TEXT("RishuProxy::on_load"),
		       TEXT("proxy->getPath failed (script)\n")
		       );
      }
      if( RishuProxy_tracelogIsEnabled(proxy) )
      {
	RishuProxy_log(proxy, TEXT("trace"),
		       TEXT("RishuProxy::on_load"),
		       TEXT("proxy->getPath failed (script)\n")
		       );
      }
      RishuProxy_myFree(proxy,path_perl);
      return FALSE;
    }
    
    /* commandLine に再構成 */
    path_perl_len = proxy->dwLoaddirLen + relative_perl_len;
    path_script_len = proxy->dwLoaddirLen + relative_script_len;
    commandLine = RishuProxy_myAlloc(proxy,(path_perl_len+path_script_len+6)*sizeof(TCHAR));
    if( !commandLine )
    {
      if( RishuProxy_errorlogIsEnabled(proxy) )
      {
	RishuProxy_log(proxy, TEXT("error"),
		       TEXT("RishuProxy::on_load"),
		       TEXT("proxy->myAlloc failed (commandLine)\n")
		       );
      }
      if( RishuProxy_tracelogIsEnabled(proxy) )
      {
	RishuProxy_log(proxy, TEXT("trace"),
		       TEXT("RishuProxy::on_load"),
		       TEXT("proxy->myAlloc failed (commandLine)\n")
		       );
      }
      RishuProxy_myFree(proxy,path_perl);
      RishuProxy_myFree(proxy,path_script);
      return FALSE;
    }
    commandLine[0] = TEXT('\"');
    CopyMemory(commandLine+1,path_perl,path_perl_len*sizeof(TCHAR));
    commandLine[path_perl_len+1] = TEXT('"');
    commandLine[path_perl_len+2] = TEXT(' ');
    commandLine[path_perl_len+3] = TEXT('"');
    CopyMemory(commandLine+path_perl_len+4,path_script,path_script_len);
    commandLine[path_perl_len+4+path_script_len] = TEXT('\"');
    commandLine[path_perl_len+4+path_script_len+1] = TEXT('\0');
    
    /* 後始末. */
    RishuProxy_myFree(proxy,path_perl);
    RishuProxy_myFree(proxy,path_script);
  }
  if( RishuProxy_systemlogIsEnabled(proxy) )
  {
    RishuProxy_log(proxy, TEXT("system"),
		   TEXT("RishuProxy::on_load"),
		   TEXT("[commandLine]\n[[\n%s\n]]\n"),
		   commandLine
		   );
  }

  /* --------------------------------------------------------------------------
   * 子プロセスの作成.
   */
  dwCreationFlags = CREATE_NO_WINDOW;
  ZeroMemory(&startupInfo,sizeof(startupInfo));
  startupInfo.cb = sizeof(startupInfo);
  startupInfo.hStdInput  = chl_rd_dup;
  startupInfo.hStdOutput = chl_wd_dup;
  startupInfo.hStdError  = chl_wd_dup;
  startupInfo.dwFlags = STARTF_USESTDHANDLES;
  res = CreateProcess(NULL,
		      commandLine,
		      NULL,
		      NULL,
		      TRUE,
		      dwCreationFlags,
		      NULL,
		      NULL,
		      &startupInfo,
		      &processInfo
		      );
  if( !res )
  {
    if( RishuProxy_errorlogIsEnabled(proxy) )
    {
      LPTSTR cause = getLastErrorMessage();
      RishuProxy_log(proxy,TEXT("error"),
		     TEXT("RishuProxy::on_load"),
		     TEXT("CreateProcess failed\n"
			  ">> %s"
			  "command-line>> %s\n"
			  ),
		     cause,
		     commandLine
		     );
      LocalFree(cause);
    }
    RishuProxy_myFree(proxy,commandLine);
    CloseHandle(par_rd);
    CloseHandle(par_wd);
    CloseHandle(chl_rd_dup);
    CloseHandle(chl_wd_dup);
    if( RishuProxy_tracelogIsEnabled(proxy) )
    {
      RishuProxy_log(proxy,TEXT("trace"),
		     TEXT("RishuProxy::on_load"),
		     TEXT("failed, CreateProcess\n")
		     );
    }
    return FALSE;
  }
  
  RishuProxy_myFree(proxy,commandLine);
  CloseHandle(chl_rd_dup);
  CloseHandle(chl_wd_dup);
  
  proxy->hChlProcess = processInfo.hProcess;
  proxy->hChlInput = par_wd;
  proxy->hChlOutput = par_rd;
  
  if( RishuProxy_tracelogIsEnabled(proxy) )
  {
    RishuProxy_log(proxy, TEXT("trace"),
		   TEXT("RishuProxy::on_load"),
		   TEXT("createProcess ok, pid = 0x%#08x\n"),
		   proxy->hChlProcess
		   );
  }
  
  /* --------------------------------------------------------------------------
   * on_load 通知.
   */
  RishuProxy_Request_init(&request,sizeof(request),0x0101);
  RishuProxy_Request_setPath(&request,"LOAD",4);
  RishuProxy_Request_write(&request,"Directory: ",11);
  RishuProxy_Request_write(&request,(LPCVOID)hGlobal_loaddir,loaddir_len);
  RishuProxy_Request_write(&request,"\r\n\r\n",4);
  RishuProxy_Request_send(&request,proxy);
  RishuProxy_Request_discard(&request);
  
  /* 回答受信 */
  RishuProxy_Response_init(&response,sizeof(response),proxy);
  ptr_response = RishuProxy_recvResponse(proxy,&response);
  res = ptr_response && (RishuProxy_Response_getStatusCode(&response)/100)==2;
  if( RishuProxy_tracelogIsEnabled(proxy) )
  {
    RishuProxy_log(proxy, TEXT("trace"),
		   TEXT("RishuProxy::on_load"),
		   TEXT("[status code]\n[[%d]]\n[response]\n"
			"[[\n%.*s]]\n[[\n%.*s]]\n"),
		   RishuProxy_Response_getStatusCode(&response),
		   response.dwHeaderLen,response.header,
		   response.dwBodyLen,response.body?response.body:""
		   );
  }
  RishuProxy_Response_discard(&response);
  
  if( RishuProxy_tracelogIsEnabled(proxy) )
  {
    RishuProxy_log(proxy, TEXT("trace"),
		   TEXT("RishuProxy::on_load"),
		   TEXT("leave [result==%s]\n"),
		   res ? "TRUE" : "FALSE"
		   );
    
  }
  return res;
}

/* ----------------------------------------------------------------------------
 * 栞 Method / request
 */
HGLOBAL RishuProxy_on_request(RishuProxy* proxy,
			      HGLOBAL hGlobal,
			      long* ptr_len
			      )
{
  RishuProxy_Request request;
  RishuProxy_Response response;
  RishuProxy_Response* ptr_response;
  
  if( !proxy )
  {
    if( hGlobal )
    {
      GlobalFree(hGlobal);
    }
    if( ptr_len )
    {
      *ptr_len = 0;
    }
    return NULL;
  }
  
  if( RishuProxy_tracelogIsEnabled(proxy) )
  {
    RishuProxy_log(proxy, TEXT("trace"),
		   TEXT("RishuProxy::on_request"),
		   TEXT("enter\n")
		   );
  }

  if( !proxy->hChlInput )
  {
    if( RishuProxy_errorlogIsEnabled(proxy) )
    {
      RishuProxy_log(proxy,TEXT("error"),
		     TEXT("RishuProxy::on_request"),
		     TEXT("not connected\n")
		     );
    }
    if( RishuProxy_tracelogIsEnabled(proxy) )
    {
      RishuProxy_log(proxy, TEXT("trace"),
		     TEXT("RishuProxy::on_request"),
		     TEXT("failure, not connected\n")
		     );
    }
    return NULL;
  }
  
  /* --------------------------------------------------------------------------
   * on_request 通知.
   */
  RishuProxy_Request_init(&request,sizeof(request),0x0101);
  RishuProxy_Request_setPath(&request,"REQUEST",7);
  RishuProxy_Request_write(&request,"Command: ",9);
  RishuProxy_Request_write(&request,(LPCVOID)hGlobal,*ptr_len);
  RishuProxy_Request_write(&request,"\r\n",2);
  RishuProxy_Request_send(&request,proxy);
  RishuProxy_Request_discard(&request);
  
  /* 回答受信 */
  RishuProxy_Response_init(&response,sizeof(response),proxy);
  ptr_response = RishuProxy_recvResponse(proxy,&response);
  if( !ptr_response )
  {
    if( RishuProxy_errorlogIsEnabled(proxy) )
    {
      RishuProxy_log(proxy,TEXT("error"),
		     TEXT("RishuProxy::on_request"),
		     TEXT("no response...\n")
		     );
    }
    RishuProxy_Response_discard(&response);
    
    hGlobal = RishuProxy_reply500(hGlobal,ptr_len);
    
    if( RishuProxy_tracelogIsEnabled(proxy) )
    {
      RishuProxy_log(proxy, TEXT("trace"),
		     TEXT("RishuProxy::on_request"),
		     TEXT("failure, no response\n")
		     );
    }
    return hGlobal;
  }
  if( (RishuProxy_Response_getStatusCode(&response)/100)!=2 )
  {
    if( RishuProxy_errorlogIsEnabled(proxy) )
    {
      RishuProxy_log(proxy,TEXT("error"),
		     TEXT("RishuProxy::on_request"),
		     TEXT("got failure [code==%d] [line==%s]\n"),
		     RishuProxy_Response_getStatusCode(&response),
		     RishuProxy_Response_getStatusLine(&response)
		     );
    }
    RishuProxy_Response_discard(&response);
    
    hGlobal = RishuProxy_reply500(hGlobal,ptr_len);
    
    if( RishuProxy_tracelogIsEnabled(proxy) )
    {
      RishuProxy_log(proxy, TEXT("trace"),
		     TEXT("RishuProxy::on_request"),
		     TEXT("failure, response tells failure\n")
		     );
    }
    return hGlobal;
  }
  
  /* hGlobalに変換. */
  hGlobal = RishuProxy_Response_intoGlobalMemory(&response,hGlobal,ptr_len);
  RishuProxy_Response_discard(&response);
  
  if( RishuProxy_tracelogIsEnabled(proxy) )
  {
    RishuProxy_log(proxy, TEXT("trace"),
		   TEXT("RishuProxy::on_request"),
		   TEXT("[leave] [hGlobal==0x%08x] [len==%d]\n"),
		   hGlobal, *ptr_len
		   );
  }
  
  return hGlobal;
}

/* ----------------------------------------------------------------------------
 * 栞 Method / unload
 */ 
BOOL RishuProxy_on_unload(RishuProxy* proxy)
{
  RishuProxy_Request request;
  RishuProxy_Response response;
  RishuProxy_Response* ptr_response;
  BOOL res;
  
  if( !proxy )
  {
    return TRUE;
  }

  if( RishuProxy_tracelogIsEnabled(proxy) )
  {
    RishuProxy_log(proxy, TEXT("trace"),
		   TEXT("RishuProxy::on_unload"),
		   TEXT("enter\n")
		   );
  }

  if( !proxy->hChlInput )
  {
    if( RishuProxy_errorlogIsEnabled(proxy) )
    {
      RishuProxy_log(proxy,TEXT("error"),
		     TEXT("RishuProxy::on_unload"),
		     TEXT("not connected\n")
		     );
    }
    if( RishuProxy_tracelogIsEnabled(proxy) )
    {
      RishuProxy_log(proxy,TEXT("trace"),
		     TEXT("RishuProxy::on_unload"),
		     TEXT("failure, not connected\n")
		     );
    }
    return FALSE;
  }
  
  /* --------------------------------------------------------------------------
   * on_unload 通知.
   */
  RishuProxy_Request_init(&request,sizeof(request),0x0101);
  RishuProxy_Request_setPath(&request,"UNLOAD",6);
  RishuProxy_Request_write(&request,"\r\n",2);
  RishuProxy_Request_send(&request,proxy);
  RishuProxy_Request_discard(&request);
  
  /* 回答受信 */
  RishuProxy_Response_init(&response,sizeof(response),proxy);
  ptr_response = RishuProxy_recvResponse(proxy,&response);
  res = ptr_response && (RishuProxy_Response_getStatusCode(&response)/100)==2;
  RishuProxy_Response_discard(&response);
  
  if( RishuProxy_tracelogIsEnabled(proxy) )
  {
    RishuProxy_log(proxy, TEXT("trace"),
		   TEXT("RishuProxy::on_unload"),
		   TEXT("leave [result==%s]\n"),
		   res ? "TRUE" : "FALSE"
		   );
    
  }
  
  return res;
}

/* ----------------------------------------------------------------------------
 * ********
 * ここからうえはチェック済み
 * ********
 */

/* ----------------------------------------------------------------------------
 * 子プロセスから受信.
 *   response は必ず一度クリアされます.
 *   エラー時には NULL を返します.
 *   エラー時には, response にエラーが起きるまでのデータは格納されますので 
 *   原因解明に役立つかも？ 
 */
RishuProxy_Response* RishuProxy_recvResponse(RishuProxy* proxy,
					     RishuProxy_Response* response
					     )
{
  LPCBYTE ptr;
  BOOL linefeed;
  DWORD endpos;
  
  /* response インスタンスのクリア. */
  if( !response )
  {
    if( RishuProxy_errorlogIsEnabled(proxy) )
    {
      RishuProxy_log(proxy,TEXT("error"),
		     TEXT("RishuProxy::recvResponse"),
		     TEXT("arg:response is required\n")
		     );
    }
    return NULL;
  }
  RishuProxy_Response_clear(response);
  
  if( !proxy )
  {
    return NULL;
  }

  /* 接続確認 */
  if( !proxy->hChlOutput )
  {
    if( RishuProxy_errorlogIsEnabled(proxy) )
    {
      RishuProxy_log(proxy,TEXT("error"),
		     TEXT("RishuProxy::recvResponse"),
		     TEXT("not connected\n")
		     );
    }
    return NULL;
  }
  
  /* 読み込みバッファの構築 */
  if( !proxy->readbuf )
  {
    proxy->readbuf = RishuProxy_myAlloc(proxy,4*1024);
    if( !proxy->readbuf )
    {
      if( RishuProxy_errorlogIsEnabled(proxy) )
      {
	RishuProxy_log(proxy,TEXT("error"),
		       TEXT("RishuProxy::recvResponse"),
		       TEXT("myAlloc failed\n")
		       );
      }
      return NULL;
    }
    proxy->dwReadBufSize = RishuProxy_mySize(proxy,proxy->readbuf);
  }
  
  /* --------------------------------------------------------------------------
   * 読み込み開始.
   */
  ptr = (LPCBYTE)proxy->readbuf;
  linefeed = proxy->dwReadBufLen==0;
  endpos = 0;
  for(;;)
  {
    DWORD readbytes;
    DWORD i;
    BOOL res;
    if( RishuProxy_tracelogIsEnabled(proxy) )
    {
      RishuProxy_log(proxy,TEXT("trace"),
		     TEXT("RishuProxy::recvResponse"),
		     TEXT("ReadFile...\n",)
		     );
    }
    res = ReadFile(proxy->hChlOutput,
		   (LPBYTE)proxy->readbuf+proxy->dwReadBufLen,
		   proxy->dwReadBufSize-proxy->dwReadBufLen,
		   &readbytes,
		   NULL
		   );
    if( !res )
    {
      if( RishuProxy_errorlogIsEnabled(proxy) )
      {
	LPTSTR cause = getLastErrorMessage();
	if( GetLastError()==ERROR_HANDLE_EOF
	    || GetLastError()==ERROR_BROKEN_PIPE )
	{
	  RishuProxy_log(proxy,TEXT("error"),
			 TEXT("RishuProxy::recvResponse"),
			 TEXT("ReadFile reach End-Of-File\n"
			      "[[\n%s]]\n"),
			 cause
			 );
	}else
        {
	  RishuProxy_log(proxy,TEXT("error"),
			 TEXT("RishuProxy::recvResponse"),
			 TEXT("ReadFile error code [%08x]\n"
			      "[[\n%s]]\n"),
			 GetLastError(),
			 cause
			 );
	}
	LocalFree(cause);
      }
      /* 読み込んでた分は追加しちゃう. */
      RishuProxy_Response_write(response,ptr,proxy->dwReadBufLen);
      proxy->dwReadBufLen = 0;
      return NULL;
    }

    if( RishuProxy_tracelogIsEnabled(proxy) )
    {
      RishuProxy_log(proxy,TEXT("trace"),
		     TEXT("RishuProxy::recvResponse"),
		     TEXT("[ReadFile]\n[[\n%.*s]]\n",),
		     readbytes,(LPCBYTE)proxy->readbuf+proxy->dwReadBufLen
		     );
    }
    proxy->dwReadBufLen += readbytes;
    
#if 0
    /* 通信内容 */
    for( i=0; i<proxy->dwReadBufLen; ++i )
    {
      register const unsigned char c = ptr[i];
      if( 0x20<=c && c<=0x7e ) putchar(c);
      else if( c=='\r' ) putchar('\\'),putchar('r');
      else if( c=='\n' ) putchar('\\'),putchar('n');
      else printf("\\x%02x",c);
    }
    putchar('\n');
#endif
    
    for( i=0; i<proxy->dwReadBufLen; ++i )
    {
      if( ptr[i]=='\r' && i+1<proxy->dwReadBufLen )
      {
	if( RishuProxy_debuglogIsEnabled(proxy) )
	{
	  RishuProxy_log(proxy,TEXT("debug"),
			 TEXT("RishuProxy::recvResponse"),
			 TEXT("(%3d) skip \\r\n"),
			 i
			 );
	}
	++i;
      }
      if( ptr[i]=='\n' )
      {
	if( linefeed )
	{ /* 空行の発見. */
	  if( RishuProxy_debuglogIsEnabled(proxy) )
	  {
	    RishuProxy_log(proxy,TEXT("debug"),
			   TEXT("RishuProxy::recvResponse"),
			   TEXT("(%3d) end-of-header found\n"),
			   i
			   );
	  }
	  endpos = i+1;
	  break;
	}else
	{ /* 行末. */
	  linefeed = TRUE;
	  if( RishuProxy_debuglogIsEnabled(proxy) )
	  {
	    RishuProxy_log(proxy,TEXT("debug"),
			   TEXT("RishuProxy::recvResponse"),
			   TEXT("(%3d) line feed found\n"),
			   i
			   );
	  }
	}
      }else if( linefeed )
      {
	if( RishuProxy_debuglogIsEnabled(proxy) )
	{
	  RishuProxy_log(proxy,TEXT("debug"),
			 TEXT("RishuProxy::recvResponse"),
			 TEXT("(%3d) cancel line feed\n"),
			 i
			 );
	}
	linefeed = FALSE;
      }
    }
    if( endpos )
    {
      RishuProxy_Response_write(response,ptr,endpos);
      proxy->dwReadBufLen -= endpos;
      if( proxy->dwReadBufLen )
      {
	MoveMemory(proxy->readbuf,ptr+endpos,proxy->dwReadBufLen);
      }
      break;
    }
    if( proxy->dwReadBufLen>2 )
    {
      RishuProxy_Response_write(response,ptr,proxy->dwReadBufLen-1);
      *(LPBYTE)proxy->readbuf = ptr[proxy->dwReadBufLen-1];
      proxy->dwReadBufLen = 1;
    }
    linefeed = FALSE;
  }
  /* ヘッダの読みとりおわり. */
  {
    LPCBYTE ptr = response->header;
    unsigned long contentLength = 0;
    const LPCBYTE tail = (LPCBYTE)response->header+response->dwHeaderLen-16;
    while( ptr<tail )
    {
      if( strnicmp(ptr,"Content-Length:",15)==0 )
      {
	char* endptr;
	ptr += 16;
	contentLength = strtoul(ptr,&endptr,0);
	if( RishuProxy_debuglogIsEnabled(proxy) )
	{
	  RishuProxy_log(proxy,TEXT("debug"),
			 TEXT("RishuProxy::recvResponse"),
			 TEXT("[found content-length] [[%.*s]]\n"),
			 endptr-ptr,ptr
			 );
	}
	break;
      }
      if( RishuProxy_debuglogIsEnabled(proxy) )
      {
	RishuProxy_log(proxy,TEXT("debug"),
		       TEXT("RishuProxy::recvResponse"),
		       TEXT("[content-length not match] [[pos=%d,%.*s]]\n"),
		       ptr-(LPBYTE)response->header,
		       15, ptr
		       );
      }
      while( *ptr++!='\n' && ptr<tail )
      {
      }
    }
    if( contentLength>0 )
    {
      DWORD ateLen;
      
      if( RishuProxy_tracelogIsEnabled(proxy) )
      {
	RishuProxy_log(proxy,TEXT("trace"),
		       TEXT("RishuProxy::recvResponse"),
		       TEXT("[content-length] [[%d]]\n"),
		       contentLength
		       );
      }
      /* 必要な分バッファの確保. */
      if( response->dwBodySize<contentLength )
      {
	RishuProxy_lock(proxy);
	if( response->body )
	{
	  response->body = RishuProxy_async_myReAlloc(proxy,response->body,contentLength);
	}else
	{
	  response->body = RishuProxy_async_myAlloc(proxy,contentLength);
	}
	if( !response->body )
	{
	  RishuProxy_unlock(proxy);
	  if( RishuProxy_tracelogIsEnabled(proxy) )
	  {
	    RishuProxy_log(proxy,TEXT("error"),
			   TEXT("RishuProxy::recvResponse"),
			   TEXT("[proxy->async_myAlloc failed]\n")
			   );
	  }
	  response->dwBodyLen = 0;
	  response->dwBodySize = 0;
	  return NULL;
	}
	response->dwBodySize = RishuProxy_async_mySize(proxy,response->body);
	RishuProxy_unlock(proxy);
      }
      /* body の読み込み〜. */
      ateLen = 0;
      for(;;)
      {
	BOOL res;
	DWORD readbytes;
	
	if( proxy->dwReadBufLen>=contentLength-ateLen )
	{ /* バッファ内に十分ある♪ */
	  DWORD eatLen = contentLength-ateLen;
	  CopyMemory((LPBYTE)response->body+ateLen,
		     proxy->readbuf,
		     eatLen
		     );
	  proxy->dwReadBufLen -= eatLen;
	  response->dwBodyLen = contentLength;
	  if( proxy->dwReadBufLen )
	  {
	    MoveMemory(proxy->readbuf,(LPCBYTE)proxy->readbuf+eatLen,proxy->dwReadBufLen);
	  }
	  break;
	}
	/* あるだけこぴー */
	CopyMemory((LPBYTE)response->body+ateLen,
		   proxy->readbuf,
		   proxy->dwReadBufLen
		   );
	ateLen += proxy->dwReadBufLen;
	proxy->dwReadBufLen = 0;
	/* もっと読み込み〜 */
	res = ReadFile(proxy->hChlOutput,
		       (LPBYTE)proxy->readbuf+proxy->dwReadBufLen,
		       proxy->dwReadBufSize-proxy->dwReadBufLen,
		       &readbytes,
		       NULL
		       );
	if( !res )
	{
	  response->dwBodyLen = ateLen;
	  if( RishuProxy_errorlogIsEnabled(proxy) )
	  {
	    LPTSTR cause = getLastErrorMessage();
	    if( GetLastError()==ERROR_HANDLE_EOF
		|| GetLastError()==ERROR_BROKEN_PIPE )
	    {
	      RishuProxy_log(proxy,TEXT("error"),
			     TEXT("RishuProxy::recvResponse"),
			     TEXT("ReadFile reach End-Of-File\n"
				  "[[\n%s]]\n"),
			     cause
			     );
	    }else
	    {
	      RishuProxy_log(proxy,TEXT("error"),
			     TEXT("RishuProxy::recvResponse"),
			     TEXT("ReadFile error code [%08x]\n"
				  "[[\n%s]]\n"),
			     GetLastError(),
			     cause
			     );
	    }
	    LocalFree(cause);
	  }
	  return NULL;
	}
	if( RishuProxy_tracelogIsEnabled(proxy) )
	{
	  RishuProxy_log(proxy,TEXT("trace"),
			 TEXT("RishuProxy::recvResponse"),
			 TEXT("[ReadFile]\n[[\n%.*s]]\n",),
			 readbytes,(LPCBYTE)proxy->readbuf+proxy->dwReadBufLen
			 );
	}
	proxy->dwReadBufLen += readbytes;
	
      }
      
    } /* contentLength>0 */
  }
  if( RishuProxy_debuglogIsEnabled(proxy) )
  {
    RishuProxy_log(proxy,TEXT("debug"),
		   TEXT("RishuProxy::recvResponse"),
		   TEXT("[succ]\n"
			"[content-length=%d]\n"
			"[content]\n"
			"[[\n%.*s]]\n"),
		   response->dwBodyLen,
		   response->dwBodyLen, response->body
		   );
  }

  return response;
}

/* ----------------------------------------------------------------------------
 * ログ (一般型) 
 */
void RishuProxy_log(RishuProxy* proxy,
		    LPCTSTR name,
		    LPCTSTR title,
		    LPCTSTR format,
		    ...
		    )
{
  va_list ap;
  va_start(ap,format);
  RishuProxy_vlog(proxy,name,title,format,ap);
  va_end(ap);
}

/* ----------------------------------------------------------------------------
 * ログ (va_list 型) 
 */
void RishuProxy_vlog(RishuProxy* proxy,
		     LPCTSTR name,
		     LPCTSTR title,
		     LPCTSTR format,
		     va_list ap
		     )
{
  if( !proxy )
  {
    return;
  }
  
  /* ログのふるい分け */
  if( _tcscmp(name,TEXT("fatal"))==0 )
  {
    if( !RishuProxy_fatallogIsEnabled(proxy) )
    {
      return;
    }
  }else if( _tcscmp(name,TEXT("error"))==0 )
  {
    if( !RishuProxy_errorlogIsEnabled(proxy) )
    {
      return;
    }
  }else if( _tcscmp(name,TEXT("warn"))==0 )
  {
    if( !RishuProxy_warnlogIsEnabled(proxy) )
    {
      return;
    }
  }else if( _tcscmp(name,TEXT("system"))==0 )
  {
    if( !RishuProxy_systemlogIsEnabled(proxy) )
    {
      return;
    }
  }else if( _tcscmp(name,TEXT("trace"))==0 )
  {
    if( !RishuProxy_tracelogIsEnabled(proxy) )
    {
      return;
    }
  }else if( _tcscmp(name,TEXT("debug"))==0 )
  {
    if( !RishuProxy_debuglogIsEnabled(proxy) )
    {
      return;
    }
  }else
  {
    return;
  }
  
  {
    /* ログ出力 */
    FILE* fp;
    if( proxy->dwLogBits & (1<<RishuProxy_LOG_CLEANUP) )
    {
      fp = _tfopen(proxy->logfilename,TEXT("wb"));
      proxy->dwLogBits &= ~(1<<RishuProxy_LOG_CLEANUP);
    }else
    {
      fp = _tfopen(proxy->logfilename,TEXT("ab"));
    }
    if( !fp )
    {
      return;
    }
    _ftprintf(fp,TEXT("<< %s (%s) >>\n"), title, name );
    _vftprintf(fp,format,ap);
    _ftprintf(fp,TEXT("--------\n"));
    fclose(fp);
  }
}

/* ----------------------------------------------------------------------------
 * makeReply
 *   回答の作成.
 */
HGLOBAL RishuProxy_makeReply(HGLOBAL hGlobal,
			     long* ptr_len,
			     LPCVOID data,
			     DWORD datalen
			     )
{
  if( (DWORD)*ptr_len < datalen && GlobalSize(hGlobal)<datalen )
  {
    hGlobal = GlobalReAlloc(hGlobal,datalen,0);
    if( !hGlobal )
    {
      *ptr_len = 0;
      return NULL;
    }
  }
  CopyMemory((LPVOID)hGlobal,data,datalen);
  *ptr_len = datalen;
  return hGlobal;
}

/* ----------------------------------------------------------------------------
 * reply500
 *   500 Internal Server Error な回答の作成.
 */
HGLOBAL RishuProxy_reply500(HGLOBAL hGlobal,
			    long* ptr_len
			    )
{
  const char* msg = 
    "SHIORI/3.0 500 Internal Server Error \r\n"
    "Sender: Rishu/Proxy\r\n"
    "\r\n";
  return RishuProxy_makeReply(hGlobal,ptr_len,msg,strlen(msg));
}
